#open"syntaxe_commune";;
#open"compilateur_commun";;

type valeur =
   Const of constante
 | Fermeture of adresse * env

and adresse == int

and env =
   Vide 
 | Cons of mutable valeur * env
;;

(* on codera le ? du let rec par la constante indefinie *)
let indefini = Const Indefini;;

type machine = {
  mutable s : adresse; (* adresse dans mem_s *)
  mutable e : env;
  mutable c : adresse; (* adresse dans mem_c *)  
  mutable d : adresse  (* adresse dans mem_d *)
};;

let TAILLE_PILE = 100000;;

let mem_s = make_vect TAILLE_PILE (Const (Int 0));;
let mem_d = make_vect TAILLE_PILE (Const (Int 0));;

let m = {s=0;e=Vide;c=0;d=0};;

let init_m () =
  m.s <-0; m.e <- Vide; m.c <- 0; m.d <-0
;;

(******************** Fonctions primitives *****************************)

let applique_nombre opi opf a b =
  match (a,b) with
    (Const (Int n1),Const (Int n2))     -> Const (Int (opi n1 n2))
  | (Const (Float n1),Const (Float n2)) -> Const (Float (opf n1 n2))
  | (Const (Int n1),Const (Float n2))   ->
         Const (Float (opf (float_of_int n1) n2))
  | (Const (Float n1),Const (Int n2))   ->
         Const (Float (opf n1 (float_of_int n2)))
  | _                                   -> failwith "Constantes non numeriques"
;;

let add a b =
  applique_nombre (prefix +) (prefix +.) a b
;;

let sub a b =
  applique_nombre (prefix -) (prefix -.) a b
;;

let mult a b =
  applique_nombre (prefix *) (prefix *.) a b
;;

let div a b =
  applique_nombre (prefix /) (prefix /.) a b
;;

let egal a b =
  Const (Bool (a = b))
;;

let affiche a =
  begin match a with
      Const (Int n) -> print_int n
    | Const (Float f) -> print_float f
    | Const (String s) -> print_string s
    | Const (Char c) -> print_char c
    | Const (Bool c) -> print_string (if c then "true" else "false")
    | Const Indefini -> print_string "_INDEFINI_"
    | Const (Nom id) -> print_string ("Symbole:("^id^")")
    | Const Unit     -> print_string "()"
    | Fermeture _    -> print_string "<fun>"
  end;
  flush std_out;
  Const Unit
;;


(************************ Execution ******************************)

let rec nth n = function
   Cons (ad,e) -> if n=1 then ad else (nth (n-1) e)
 | Vide        -> failwith"anomalie_nth"
;;

let rec mod_nth n v = function
   Cons ((_,e) as ad) -> if n=1 then ad <- (v,e) else (mod_nth (n-1) v e)
 | Vide               -> failwith"anomalie_mod_nth"
;;

let tail = function
   Cons (_,e) -> e
 | Vide       -> failwith"anomalie_tail"
;;

exception Fini;;

let execute1 mem_c = match (mem_c.(m.c)) with
   SAccede n         -> 
     m.s <- m.s + 1;
     mem_s.(m.s) <- (nth n m.e);
     m.c <- m.c + 1
 | SApplique         ->
     m.d <- m.d + 1;
     mem_d.(m.d) <- Fermeture (m.c + 1,m.e);
     begin match mem_s.(m.s) with
       Const _           -> failwith "anomalie2"
     | Fermeture (c',e') -> m.s <- m.s - 1;
                            m.c <- c';
                            m.e <- e'
     end
 | SConstante cst    ->
     m.s <- m.s + 1;
     mem_s.(m.s) <- Const cst;
     m.c <- m.c + 1
 | SFonction ad      ->
     m.s <- m.s + 1;
     mem_s.(m.s) <- Fermeture (ad,m.e);
     m.c <- m.c + 1
 | SLambda           ->
     m.e <- Cons (mem_s.(m.s),m.e);
     m.s <- m.s - 1;
     m.c <- m.c + 1
 | SDepile n         ->
     m.c <- m.c + 1;
     for i=1 to n do
       m.e <- tail m.e
     done
 | SRetour           ->
     if m.d = 0 then raise Fini;
     begin match mem_d.(m.d) with
       Const _           -> failwith "anomalie3"
     | Fermeture (c',e') -> m.c <- c';m.e <- e'
     end;
     m.d <- m.d - 1
 | SReserve n        ->
     for i=1 to n do
       m.e <- Cons (indefini ,m.e)
     done;
     m.c <- m.c + 1
 | SLet n            ->
     for i=0 to n-1 do
       m.e <- Cons (mem_s.(m.s - i),m.e)
     done;
     m.s <- m.s - n;
     m.c <- m.c + 1
 | SLetrec n         ->
     for i=1 to n do
       mod_nth i mem_s.(m.s - i + 1) m.e
     done;
     m.s <- m.s - n;
     m.c <- m.c + 1
 | SSelect (ad1,ad2) ->
     m.d <- m.d + 1;
     mem_d.(m.d) <- Fermeture (m.c + 1,m.e);
     if mem_s.(m.s)=Const (Bool false) then
	   (m.s <- m.s - 1; m.c <- ad2)
     else
	   (m.s <- m.s - 1; m.c <- ad1)
 | SPrimitive op     -> 
     begin match op with
       Plus    -> m.s <- m.s + 1; mem_s.(m.s) <- add (nth 2 m.e) (nth 1 m.e)
     | Moins   -> m.s <- m.s + 1; mem_s.(m.s) <- sub (nth 2 m.e) (nth 1 m.e)
     | Fois    -> m.s <- m.s + 1; mem_s.(m.s) <- mult (nth 2 m.e) (nth 1 m.e)
     | Divise  -> m.s <- m.s + 1; mem_s.(m.s) <- div (nth 2 m.e) (nth 1 m.e)
     | Egal    -> m.s <- m.s + 1; mem_s.(m.s) <- egal (nth 2 m.e) (nth 1 m.e)
     | Affiche -> m.s <- m.s + 1; mem_s.(m.s) <- affiche (nth 1 m.e)
     end;
     if m.d = 0 then raise Fini;
     begin match mem_d.(m.d) with
       Const _           -> failwith "anomalie4"
     | Fermeture (c',e') -> m.c <- c';m.e <- e'
     end;
     m.d <- m.d - 1
 | SModifie n        ->
     mod_nth n mem_s.(m.s) m.e;
     m.c <- m.c + 1
;;

let execute fic =
  let canal = open_in fic in
  input_line canal;
  let mem_c = (input_value canal : seqasm vect) in
  init_m ();
  try
    while true do
(* TRACE DE L'EXECUTION
      print_int m.s;
      print_string " ";
      print_int m.c;
      print_string " ";
      print_int m.d;
      print_newline ();
*)
      execute1 mem_c
    done
  with Fini -> ()
     | Failure s -> print_string ("Erreur: "^s^"\n");print_newline ()
;;

if not sys__interactive then
   execute sys__command_line.(1)
;;

(*
include"charge_interprete";;
execute "../EXEMPLES-MINIML/exemple1";;
*)
